@@ -2,37 +2,32 @@
 the flask extension
 """
 import datetime
+import enum
 import itertools
 import logging
 import time
-import warnings
-
 from functools import wraps
+from typing import Callable, Dict, List, Optional, Set, Tuple, Union, cast
 
-from typing import Callable
-from typing import Dict
-from typing import List
-from typing import Optional
-from typing import Set
-from typing import Union
-
-from flask import request, current_app, g, Blueprint, Flask, Response
+from flask import Blueprint, Flask, Response, current_app, g, request
+from limits import RateLimitItem
 from limits.errors import ConfigurationError
-from limits.storage import storage_from_string, Storage, MemoryStorage
+from limits.storage import MemoryStorage, Storage, storage_from_string
 from limits.strategies import STRATEGIES, RateLimiter
 from werkzeug.http import http_date, parse_date
 
 from flask_limiter.wrappers import Limit, LimitGroup
+
 from .errors import RateLimitExceeded
 
 
 class C:
     ENABLED = "RATELIMIT_ENABLED"
     HEADERS_ENABLED = "RATELIMIT_HEADERS_ENABLED"
-    STORAGE_URL = "RATELIMIT_STORAGE_URL"
+    STORAGE_URI = "RATELIMIT_STORAGE_URI"
+    STORAGE_URL = "RATELIMIT_STORAGE_URL"  # Deprecated due to inconsistency.
     STORAGE_OPTIONS = "RATELIMIT_STORAGE_OPTIONS"
     STRATEGY = "RATELIMIT_STRATEGY"
-    GLOBAL_LIMITS = "RATELIMIT_GLOBAL"
     DEFAULT_LIMITS = "RATELIMIT_DEFAULT"
     DEFAULT_LIMITS_PER_METHOD = "RATELIMIT_DEFAULTS_PER_METHOD"
     DEFAULT_LIMITS_EXEMPT_WHEN = "RATELIMIT_DEFAULTS_EXEMPT_WHEN"
@@ -47,13 +42,71 @@ class C:
     HEADER_RETRY_AFTER = "RATELIMIT_HEADER_RETRY_AFTER"
     HEADER_RETRY_AFTER_VALUE = "RATELIMIT_HEADER_RETRY_AFTER_VALUE"
     KEY_PREFIX = "RATELIMIT_KEY_PREFIX"
+    FAIL_ON_FIRST_BREACH = "RATELIMIT_FAIL_ON_FIRST_BREACH"
 
 
-class HEADERS:
-    RESET = 1
-    REMAINING = 2
-    LIMIT = 3
-    RETRY_AFTER = 4
+class HEADERS(enum.Enum):
+    """
+    Enumeration of supported rate limit related headers to
+    be used when configuring via :paramref:`~flask_limiter.Limiter.header_name_mapping`
+    """
+
+    #: Timestamp at which this rate limit will be reset
+    RESET = "X-RateLimit-Reset"
+    #: Remaining number of requests within the current window
+    REMAINING = "X-RateLimit-Remaining"
+    #: Total number of allowed requests within a window
+    LIMIT = "X-RateLimit-Limit"
+    #: Number of seconds to retry after at
+    RETRY_AFTER = "Retry-After"
+
+
+class RequestLimit:
+    """
+    Provides details of a rate limit within the context of a request
+    """
+
+    #: The instance of the rate limit
+    limit: RateLimitItem
+
+    #: The full key for the request against which the rate limit is tested
+    key: str
+
+    #: Whether the limit was breached within the context of this request
+    breached: bool
+
+    def __init__(
+        self,
+        limiter: RateLimiter,
+        limit: RateLimitItem,
+        request_args: List[str],
+        breached: bool,
+    ):
+        self.limiter = limiter
+        self.limit = limit
+        self.request_args = request_args
+        self.key = limit.key_for(*request_args)
+        self.breached = breached
+        self._window: Optional[Tuple[int, int]] = None
+
+    @property
+    def window(self):
+        if not self._window:
+            self._window = self.limiter.get_window_stats(self.limit, *self.request_args)
+
+        return self._window
+
+    @property
+    def reset_at(self) -> int:
+        """Timestamp at which the rate limit will be reset"""
+
+        return int(self.window[0] + 1)
+
+    @property
+    def remaining(self) -> int:
+        """Quantity remaining for this rate limit"""
+
+        return self.window[1]
 
 
 MAX_BACKEND_CHECKS = 5
@@ -67,7 +120,7 @@ class Limiter(object):
     :param key_func: a callable that returns the domain to rate limit
       by.
     :param default_limits: a variable list of strings or callables
-     returning strings denoting global limits to apply to all routes.
+     returning strings denoting default limits to apply to all routes.
      :ref:`ratelimit-string` for  more details.
     :param default_limits_per_method: whether default limits are applied
      per method, per route or as a combination of all method per route.
@@ -81,21 +134,26 @@ class Limiter(object):
      (i.e a shared limit for all routes)
     :param headers_enabled: whether ``X-RateLimit`` response headers are
      written.
-    :param strategy: the strategy to use.
-     Refer to :ref:`ratelimit-strategy`
+    :param header_name_mapping: Mapping of header names to use if
+     :paramref:`Limiter.headers_enabled` is ``True``. If no mapping is provided
+     the default values will be used.
+    :param strategy: the strategy to use. Refer to :ref:`ratelimit-strategy`
     :param storage_uri: the storage location.
-     Refer to :ref:`ratelimit-conf`
+     Refer to :data:`RATELIMIT_STORAGE_URI`
     :param storage_options: kwargs to pass to the storage implementation
      upon instantiation.
     :param auto_check: whether to automatically check the rate limit in
      the before_request chain of the application. default ``True``
     :param swallow_errors: whether to swallow errors when hitting a rate
      limit. An exception will still be logged. default ``False``
+    :param fail_on_first_breach: whether to stop processing remaining limits
+     after the first breach. default ``True``
     :param in_memory_fallback: a variable list of strings or callables
      returning strings denoting fallback limits to apply when the storage is
      down.
-    :param in_memory_fallback_enabled: simply falls back to in memory
+    :param in_memory_fallback_enabled: fall back to in memory
      storage when the main storage is down and inherits the original limits.
+     default ``False``
     :param retry_after: Allows configuration of how the value of the
      `Retry-After` header is rendered. One of `http-date` or `delta-seconds`.
     :param key_prefix: prefix prepended to rate limiter keys and app context global names.
@@ -106,20 +164,21 @@ class Limiter(object):
         self,
         app: Optional[Flask] = None,
         key_func: Callable[[], str] = None,
-        global_limits: List[str] = [],
-        default_limits: List[str] = [],
-        default_limits_per_method=False,
+        default_limits: List[Union[str, Callable[[], str]]] = [],
+        default_limits_per_method: bool = None,
         default_limits_exempt_when: Callable[[], bool] = None,
         default_limits_deduct_when: Callable[[], bool] = None,
-        application_limits: List[str] = [],
-        headers_enabled=False,
+        application_limits: List[Union[str, Callable[[], str]]] = [],
+        headers_enabled: bool = None,
+        header_name_mapping: Dict[HEADERS, str] = {},
         strategy: Optional[str] = None,
         storage_uri: Optional[str] = None,
         storage_options={},
-        auto_check=True,
-        swallow_errors=False,
+        auto_check: bool = True,
+        swallow_errors: bool = None,
+        fail_on_first_breach: bool = None,
         in_memory_fallback: List[str] = [],
-        in_memory_fallback_enabled=False,
+        in_memory_fallback_enabled: bool = None,
         retry_after: str = None,
         key_prefix: str = "",
         enabled: bool = True,
@@ -141,28 +200,36 @@ class Limiter(object):
         self._exempt_routes: Set[str] = set()
         self._request_filters: List[Callable[[], bool]] = []
         self._headers_enabled = headers_enabled
-        self._header_mapping: Dict[int, str] = {}
+        self._header_mapping = header_name_mapping
         self._retry_after = retry_after
         self._strategy = strategy
         self._storage_uri = storage_uri
         self._storage_options = storage_options
         self._auto_check = auto_check
         self._swallow_errors = swallow_errors
+        self._fail_on_first_breach = fail_on_first_breach
 
         # No longer optional
         assert key_func
 
-        if global_limits:
-            self.__raise_global_limits_warning()
-
         self._key_func = key_func
         self._key_prefix = key_prefix
 
-        for limit in set(global_limits + default_limits):
+        for limit in default_limits:
             self._default_limits.extend(
                 [
                     LimitGroup(
-                        limit, self._key_func, None, False, None, None, None, None, None
+                        limit,
+                        self._key_func,
+                        None,
+                        False,
+                        None,
+                        None,
+                        None,
+                        None,
+                        None,
+                        None,
+                        1,
                     )
                 ]
             )
@@ -180,6 +247,8 @@ class Limiter(object):
                         None,
                         None,
                         None,
+                        None,
+                        1,
                     )
                 ]
             )
@@ -188,7 +257,17 @@ class Limiter(object):
             self._in_memory_fallback.extend(
                 [
                     LimitGroup(
-                        limit, self._key_func, None, False, None, None, None, None, None
+                        limit,
+                        self._key_func,
+                        None,
+                        False,
+                        None,
+                        None,
+                        None,
+                        None,
+                        None,
+                        None,
+                        1,
                     )
                 ]
             )
@@ -224,23 +303,35 @@ class Limiter(object):
         if not self.enabled:
             return
 
-        self._default_limits_per_method = config.setdefault(
-            C.DEFAULT_LIMITS_PER_METHOD, self._default_limits_per_method
+        if self._default_limits_per_method is None:
+            self._default_limits_per_method = config.get(
+                C.DEFAULT_LIMITS_PER_METHOD, False
+            )
+        self._default_limits_exempt_when = (
+            self._default_limits_exempt_when or config.get(C.DEFAULT_LIMITS_EXEMPT_WHEN)
         )
-        self._default_limits_exempt_when = config.setdefault(
-            C.DEFAULT_LIMITS_EXEMPT_WHEN, self._default_limits_exempt_when
-        )
-        self._default_limits_deduct_when = config.setdefault(
-            C.DEFAULT_LIMITS_DEDUCT_WHEN, self._default_limits_deduct_when
-        )
-        self._swallow_errors = config.setdefault(C.SWALLOW_ERRORS, self._swallow_errors)
-        self._headers_enabled = self._headers_enabled or config.setdefault(
-            C.HEADERS_ENABLED, False
+        self._default_limits_deduct_when = (
+            self._default_limits_deduct_when or config.get(C.DEFAULT_LIMITS_DEDUCT_WHEN)
         )
+
+        if self._swallow_errors is None:
+            self._swallow_errors = config.get(C.SWALLOW_ERRORS, False)
+
+        if self._fail_on_first_breach is None:
+            self._fail_on_first_breach = config.get(C.FAIL_ON_FIRST_BREACH, True)
+
+        if self._headers_enabled is None:
+            self._headers_enabled = config.get(C.HEADERS_ENABLED, False)
+
         self._storage_options.update(config.get(C.STORAGE_OPTIONS, {}))
-        self._storage = storage_from_string(
-            self._storage_uri or config.setdefault(C.STORAGE_URL, "memory://"),
-            **self._storage_options
+        storage_uri_from_config = config.get(
+            C.STORAGE_URI, config.get(C.STORAGE_URL, "memory://")
+        )
+        self._storage = cast(
+            Storage,
+            storage_from_string(
+                self._storage_uri or storage_uri_from_config, **self._storage_options
+            ),
         )
         strategy = self._strategy or config.setdefault(C.STRATEGY, "fixed-window")
 
@@ -248,21 +339,20 @@ class Limiter(object):
             raise ConfigurationError("Invalid rate limiting strategy %s" % strategy)
         self._limiter = STRATEGIES[strategy](self._storage)
 
-        # TODO: this should be made consistent with the rest of the
-        #  configuration
         self._header_mapping = {
             HEADERS.RESET: self._header_mapping.get(
-                HEADERS.RESET, config.get(C.HEADER_RESET, "X-RateLimit-Reset")
+                HEADERS.RESET, config.get(C.HEADER_RESET, HEADERS.RESET.value)
             ),
             HEADERS.REMAINING: self._header_mapping.get(
                 HEADERS.REMAINING,
-                config.get(C.HEADER_REMAINING, "X-RateLimit-Remaining"),
+                config.get(C.HEADER_REMAINING, HEADERS.REMAINING.value),
             ),
             HEADERS.LIMIT: self._header_mapping.get(
-                HEADERS.LIMIT, config.get(C.HEADER_LIMIT, "X-RateLimit-Limit")
+                HEADERS.LIMIT, config.get(C.HEADER_LIMIT, HEADERS.LIMIT.value)
             ),
             HEADERS.RETRY_AFTER: self._header_mapping.get(
-                HEADERS.RETRY_AFTER, config.get(C.HEADER_RETRY_AFTER, "Retry-After")
+                HEADERS.RETRY_AFTER,
+                config.get(C.HEADER_RETRY_AFTER, HEADERS.RETRY_AFTER.value),
             ),
         }
         self._retry_after = self._retry_after or config.get(C.HEADER_RETRY_AFTER_VALUE)
@@ -283,13 +373,12 @@ class Limiter(object):
                     None,
                     None,
                     None,
+                    None,
+                    1,
                 )
             ]
 
-        if config.get(C.GLOBAL_LIMITS, None):
-            self.__raise_global_limits_warning()
-
-        conf_limits = config.get(C.GLOBAL_LIMITS, config.get(C.DEFAULT_LIMITS, None))
+        conf_limits = config.get(C.DEFAULT_LIMITS, None)
 
         if not self._default_limits and conf_limits:
             self._default_limits = [
@@ -303,6 +392,8 @@ class Limiter(object):
                     None,
                     None,
                     None,
+                    None,
+                    1,
                 )
             ]
 
@@ -326,6 +417,134 @@ class Limiter(object):
         app.extensions["limiter"] = self
         self.initialized = True
 
+    def limit(
+        self,
+        limit_value: Union[str, Callable[[], str]],
+        key_func: Callable[[], str] = None,
+        per_method: bool = False,
+        methods: List[str] = None,
+        error_message: str = None,
+        exempt_when: Callable[[], bool] = None,
+        override_defaults: bool = True,
+        deduct_when: Callable[[Response], bool] = None,
+        on_breach: Callable[[RequestLimit], None] = None,
+        cost: int = 1,
+    ) -> Callable:
+        """
+        decorator to be used for rate limiting individual routes or blueprints.
+
+        :param limit_value: rate limit string or a callable that returns a
+         string. :ref:`ratelimit-string` for more details.
+        :param key_func: function/lambda to extract the unique
+         identifier for the rate limit. defaults to remote address of the
+         request.
+        :param per_method: whether the limit is sub categorized into the
+         http method of the request.
+        :param methods: if specified, only the methods in this list will
+         be rate limited (default: None).
+        :param error_message: string (or callable that returns one) to override
+         the error message used in the response.
+        :param exempt_when: function/lambda used to decide if the rate
+         limit should skipped.
+        :param override_defaults:  whether the decorated limit overrides
+         the default limits. (default: True)
+        :param deduct_when: a function that receives the current
+         :class:`flask.Response` object and returns True/False to decide if a
+         deduction should be done from the rate limit
+        :param on_breach: a function that will be called when this limit
+         is breached.
+        :param cost: cost of a hit (default: 1).
+        """
+
+        return self.__limit_decorator(
+            limit_value,
+            key_func,
+            per_method=per_method,
+            methods=methods,
+            error_message=error_message,
+            exempt_when=exempt_when,
+            override_defaults=override_defaults,
+            deduct_when=deduct_when,
+            on_breach=on_breach,
+            cost=cost,
+        )
+
+    def shared_limit(
+        self,
+        limit_value: Union[str, Callable[[], str]],
+        scope: Union[str, Callable[[], str]],
+        key_func: Callable[[], str] = None,
+        error_message: str = None,
+        exempt_when: Callable[[], bool] = None,
+        override_defaults=True,
+        deduct_when: Callable[[Response], bool] = None,
+        on_breach: Callable[[RequestLimit], None] = None,
+        cost: int = 1,
+    ) -> Callable:
+        """
+        decorator to be applied to multiple routes sharing the same rate limit.
+
+        :param limit_value: rate limit string or a callable that returns a
+         string. :ref:`ratelimit-string` for more details.
+        :param scope: a string or callable that returns a string
+         for defining the rate limiting scope.
+        :param key_func: function/lambda to extract the unique
+         identifier for the rate limit. defaults to remote address of the
+         request.
+        :param error_message: string (or callable that returns one) to override
+         the error message used in the response.
+        :param function exempt_when: function/lambda used to decide if the rate
+         limit should skipped.
+        :param override_defaults:  whether the decorated limit overrides
+         the default limits. (default: True)
+        :param deduct_when: a function that receives the current
+         :class:`flask.Response`  object and returns True/False to decide if a
+         deduction should be done from the rate limit
+        :param on_breach: a function that will be called when this limit
+         is breached.
+        :param cost: cost of a hit (default: 1).
+        """
+
+        return self.__limit_decorator(
+            limit_value,
+            key_func,
+            True,
+            scope,
+            error_message=error_message,
+            exempt_when=exempt_when,
+            override_defaults=override_defaults,
+            deduct_when=deduct_when,
+            on_breach=on_breach,
+            cost=cost,
+        )
+
+    def exempt(self, obj: Union[Callable, Blueprint]):
+        """
+        decorator to mark a view or all views in a blueprint as exempt from
+        rate limits.
+
+        :param obj: view or blueprint to mark as exempt.
+        """
+
+        if isinstance(obj, Blueprint):
+            self._blueprint_exempt.add(obj.name)
+        else:
+            self._exempt_routes.add(f"{obj.__module__}.{obj.__name__}")
+
+            return obj
+
+    def request_filter(self, fn: Callable[[], bool]) -> Callable:
+        """
+        decorator to mark a function as a filter to be executed
+        to check if the request is exempt from rate limiting.
+
+        :param fn: The function will be called before evaluating any rate limits
+         to decide whether to perform rate limit or skip it.
+        """
+        self._request_filters.append(fn)
+
+        return fn
+
     def __configure_fallbacks(self, app, strategy):
         config = app.config
         fallback_enabled = config.get(C.IN_MEMORY_FALLBACK_ENABLED, False)
@@ -343,6 +562,8 @@ class Limiter(object):
                     None,
                     None,
                     None,
+                    None,
+                    1,
                 )
             ]
 
@@ -392,31 +613,90 @@ class Limiter(object):
         else:
             return self._limiter
 
+    @property
+    def current_limit(self) -> Optional[RequestLimit]:
+        """
+        Get details for the most relevant rate limit used in this request.
+
+        In a scenario where multiple rate limits are active for a single request
+        and none are breached, the rate limit which applies to the smallest
+        time window will be returned.
+
+        .. important:: The value of ``remaining`` in :class:`RequestLimit` is after
+           deduction for the current request.
+
+
+        For example::
+
+            @limit("1/second")
+            @limit("60/minute")
+            @limit("2/day")
+            def route(...):
+                ...
+
+        - Request 1 at ``t=0`` (no breach): this will return the details for for ``1/second``
+        - Request 2 at ``t=1`` (no breach): it will still return the details for ``1/second``
+        - Request 3 at ``t=2`` (breach): it will return the details for ``2/day``
+        """
+        last_limit = getattr(g, f"{self._key_prefix}_view_rate_limit", None)
+        breached_limits = getattr(g, f"{self._key_prefix}_breached_limits", [])
+
+        if last_limit:
+            return RequestLimit(
+                limit=last_limit[0],
+                limiter=self.limiter,
+                request_args=last_limit[1:],
+                breached=last_limit[0] in breached_limits,
+            )
+
+        return None
+
+    @property
+    def current_limits(self) -> List[RequestLimit]:
+        """
+        Get a list of all rate limits that were applicable and evaluated
+        within the context of this request.
+
+        The limits are returned in a sorted order by smallest window size first.
+        """
+        all_limits = getattr(g, f"{self._key_prefix}_view_rate_limits", [])
+        breached_limits = getattr(g, f"{self._key_prefix}_breached_limits", [])
+
+        return list(
+            RequestLimit(
+                limit=limit[0],
+                limiter=self.limiter,
+                request_args=limit[1:],
+                breached=limit[0] in breached_limits,
+            )
+            for limit in sorted(all_limits)
+        )
+
     def __check_conditional_deductions(self, response):
 
         for lim, args in getattr(
-            g, "%s_conditional_deductions" % self._key_prefix, {}
+            g, f"{self._key_prefix}_conditional_deductions", {}
         ).items():
             if lim.deduct_when(response):
-                self.limiter.hit(lim.limit, *args)
+                self.limiter.hit(lim.limit, *args, cost=lim.cost)
 
         return response
 
     def __inject_headers(self, response):
         self.__check_conditional_deductions(response)
-        current_limit = getattr(g, "%s_view_rate_limit" % self._key_prefix, None)
+        header_limit = self.current_limit
 
-        if self.enabled and self._headers_enabled and current_limit:
+        if self.enabled and self._headers_enabled and header_limit:
             try:
-                window_stats = self.limiter.get_window_stats(*current_limit)
-                reset_in: Union[int, float] = 1 + window_stats[0]
+                reset_at = header_limit.reset_at
                 response.headers.add(
-                    self._header_mapping[HEADERS.LIMIT], str(current_limit[0].amount)
+                    self._header_mapping[HEADERS.LIMIT],
+                    str(header_limit.limit.amount),
                 )
                 response.headers.add(
-                    self._header_mapping[HEADERS.REMAINING], window_stats[1]
+                    self._header_mapping[HEADERS.REMAINING], header_limit.remaining
                 )
-                response.headers.add(self._header_mapping[HEADERS.RESET], reset_in)
+                response.headers.add(self._header_mapping[HEADERS.RESET], reset_at)
 
                 # response may have an existing retry after
                 existing_retry_after_header = response.headers.get("Retry-After")
@@ -435,14 +715,14 @@ class Limiter(object):
                     if isinstance(retry_after, datetime.datetime):
                         retry_after = time.mktime(retry_after.timetuple())
 
-                    reset_in = max(retry_after, reset_in)
+                    reset_at = max(int(retry_after), reset_at)
 
                 # set the header instead of using add
                 response.headers.set(
                     self._header_mapping[HEADERS.RETRY_AFTER],
                     self._retry_after == "http-date"
-                    and http_date(reset_in)
-                    or int(reset_in - time.time()),
+                    and http_date(reset_at)
+                    or int(reset_at - time.time()),
                 )
             except Exception as e:  # noqa: E722
                 if self._in_memory_fallback_enabled and not self._storage_dead:
@@ -463,14 +743,14 @@ class Limiter(object):
         return response
 
     def __evaluate_limits(self, endpoint, limits):
-        failed_limit = None
+        failed_limits = []
         limit_for_header = None
-        limits_escape = []
+        view_limits = []
 
         if not getattr(g, "%s_conditional_deductions" % self._key_prefix, None):
             setattr(g, "%s_conditional_deductions" % self._key_prefix, {})
 
-        for lim in limits:
+        for lim in sorted(limits, key=lambda x: x.limit):
             limit_scope = lim.scope or endpoint
 
             if lim.is_exempt or lim.method_exempt:
@@ -480,6 +760,7 @@ class Limiter(object):
                 limit_scope += ":%s" % request.method
             limit_key = lim.key_func()
             args = [limit_key, limit_scope]
+            kwargs = {}
 
             if not all(args):
                 self.logger.error(
@@ -496,34 +777,57 @@ class Limiter(object):
                 method = self.limiter.test
             else:
                 method = self.limiter.hit
+                kwargs["cost"] = lim.cost
 
             if not limit_for_header or lim.limit < limit_for_header[0]:
                 limit_for_header = [lim.limit] + args
 
-            if not method(lim.limit, *args):
+            view_limits.append([lim.limit] + args)
+
+            if not method(lim.limit, *args, **kwargs):
                 self.logger.warning(
                     "ratelimit %s (%s) exceeded at endpoint: %s",
                     lim.limit,
                     limit_key,
                     limit_scope,
                 )
-                failed_limit = lim
+                failed_limits.append([lim, args])
                 limit_for_header = [lim.limit] + args
 
-                break
+                if self._fail_on_first_breach:
+                    break
 
-            limits_escape.append([lim.limit] + args)
+        setattr(g, f"{self._key_prefix}_view_rate_limit", limit_for_header)
+        setattr(g, f"{self._key_prefix}_view_rate_limits", view_limits)
 
-        setattr(g, "%s_view_rate_limit" % self._key_prefix, limit_for_header)
-        setattr(g, "%s_view_limits" % self._key_prefix, limits_escape)
+        if failed_limits:
+            inner_limits = [limit[0] for limit in failed_limits]
 
-        if failed_limit:
-            raise RateLimitExceeded(failed_limit)
+            for limit in failed_limits:
+                if limit[0].on_breach:
+                    try:
+                        limit[0].on_breach(
+                            RequestLimit(
+                                self.limiter,
+                                limit=limit[0].limit,
+                                request_args=limit[1:],
+                                breached=True,
+                            )
+                        )
+                    except Exception:  # noqa
+                        self.logger.warning("on_breach callback failed")
+
+            setattr(
+                g,
+                f"{self._key_prefix}_breached_limits",
+                [limit.limit for limit in inner_limits],
+            )
+            raise RateLimitExceeded(sorted(inner_limits)[0])
 
     def __check_request_limit(self, in_middleware=True):
         endpoint = request.endpoint or ""
         view_func = current_app.view_functions.get(endpoint, None)
-        name = "%s.%s" % (view_func.__module__, view_func.__name__) if view_func else ""
+        name = f"{view_func.__module__}.{view_func.__name__}" if view_func else ""
 
         if (
             not request.endpoint
@@ -532,31 +836,13 @@ class Limiter(object):
             or name in self._exempt_routes
             or request.blueprint in self._blueprint_exempt
             or any(fn() for fn in self._request_filters)
-            or g.get("%s_rate_limiting_complete" % self._key_prefix)
+            or g.get(f"{self._key_prefix}_rate_limiting_complete")
         ):
             return
         limits: List[Limit] = []
         dynamic_limits: List[Limit] = []
 
-        # this is to ensure backward compatibility with behavior that
-        # existed accidentally, i.e::
-        #
-        # @limiter.limit(...)
-        # @app.route('...')
-        # def func(...):
-        #
-        # The above setup would work in pre 1.0 versions because the decorator
-        # was not acting immediately and instead simply registering the rate
-        # limiting. The correct way to use the decorator is to wrap
-        # the limiter with the route, i.e::
-        #
-        # @app.route(...)
-        # @limiter.limit(...)
-        # def func(...):
-
-        implicit_decorator = view_func in self.__marked_for_limiting.get(name, [])
-
-        if not in_middleware or implicit_decorator:
+        if not in_middleware:
             limits = name in self._route_limits and self._route_limits[name] or []
             dynamic_limits = []
 
@@ -590,6 +876,8 @@ class Limiter(object):
                                     limit.exempt_when,
                                     limit.override_defaults,
                                     limit.deduct_when,
+                                    limit.on_breach,
+                                    limit.cost,
                                 )
                                 for limit in limit_group
                             ]
@@ -637,10 +925,8 @@ class Limiter(object):
                 )
 
                 if (
-                    (explicit_limits_exempt or combined_defaults)
-                    and not before_request_context
-                    or implicit_decorator
-                ):
+                    explicit_limits_exempt or combined_defaults
+                ) and not before_request_context:
                     all_limits += list(itertools.chain(*self._default_limits))
             self.__evaluate_limits(endpoint, all_limits)
         except Exception as e:
@@ -672,13 +958,15 @@ class Limiter(object):
         exempt_when=None,
         override_defaults=True,
         deduct_when=None,
+        on_breach=None,
+        cost=1,
     ):
         _scope = scope if shared else None
 
         def _inner(obj):
             func = key_func or self._key_func
             is_route = not isinstance(obj, Blueprint)
-            name = "%s.%s" % (obj.__module__, obj.__name__) if is_route else obj.name
+            name = f"{obj.__module__}.{obj.__name__}" if is_route else obj.name
             dynamic_limit, static_limits = None, []
 
             if callable(limit_value):
@@ -692,6 +980,8 @@ class Limiter(object):
                     exempt_when,
                     override_defaults,
                     deduct_when,
+                    on_breach,
+                    cost,
                 )
             else:
                 try:
@@ -706,6 +996,8 @@ class Limiter(object):
                             exempt_when,
                             override_defaults,
                             deduct_when,
+                            on_breach,
+                            cost,
                         )
                     )
                 except ValueError as e:
@@ -736,138 +1028,13 @@ class Limiter(object):
                 @wraps(obj)
                 def __inner(*a, **k):
                     if self._auto_check and not g.get(
-                        "%s_rate_limiting_complete" % self._key_prefix
+                        f"{self._key_prefix}_rate_limiting_complete"
                     ):
                         self.__check_request_limit(False)
-                        setattr(g, "%s_rate_limiting_complete" % self._key_prefix, True)
+                        setattr(g, f"{self._key_prefix}_rate_limiting_complete", True)
 
                     return current_app.ensure_sync(obj)(*a, **k)
 
                 return __inner
 
         return _inner
-
-    def limit(
-        self,
-        limit_value: Union[str, Callable[[], str]],
-        key_func: Callable[[], str] = None,
-        per_method: bool = False,
-        methods: List[str] = None,
-        error_message: str = None,
-        exempt_when: Callable[[], bool] = None,
-        override_defaults: bool = True,
-        deduct_when: Callable[[Response], bool] = None,
-    ) -> Callable:
-        """
-        decorator to be used for rate limiting individual routes or blueprints.
-
-        :param limit_value: rate limit string or a callable that returns a
-         string. :ref:`ratelimit-string` for more details.
-        :param key_func: function/lambda to extract the unique
-         identifier for the rate limit. defaults to remote address of the
-         request.
-        :param per_method: whether the limit is sub categorized into the
-         http method of the request.
-        :param methods: if specified, only the methods in this list will
-         be rate limited (default: None).
-        :param error_message: string (or callable that returns one) to override
-         the error message used in the response.
-        :param exempt_when: function/lambda used to decide if the rate
-         limit should skipped.
-        :param override_defaults:  whether the decorated limit overrides
-         the default limits. (default: True)
-        :param deduct_when: a function that receives the current
-         :class:`flask.Response` object and returns True/False to decide if a
-         deduction should be done from the rate limit
-        """
-
-        return self.__limit_decorator(
-            limit_value,
-            key_func,
-            per_method=per_method,
-            methods=methods,
-            error_message=error_message,
-            exempt_when=exempt_when,
-            override_defaults=override_defaults,
-            deduct_when=deduct_when,
-        )
-
-    def shared_limit(
-        self,
-        limit_value: Union[str, Callable[[], str]],
-        scope: Union[str, Callable[[], str]],
-        key_func: Callable[[], str] = None,
-        error_message: str = None,
-        exempt_when: Callable[[], bool] = None,
-        override_defaults=True,
-        deduct_when: Callable[[Response], bool] = None,
-    ) -> Callable:
-        """
-        decorator to be applied to multiple routes sharing the same rate limit.
-
-        :param limit_value: rate limit string or a callable that returns a
-         string. :ref:`ratelimit-string` for more details.
-        :param scope: a string or callable that returns a string
-         for defining the rate limiting scope.
-        :param key_func: function/lambda to extract the unique
-         identifier for the rate limit. defaults to remote address of the
-         request.
-        :param error_message: string (or callable that returns one) to override
-         the error message used in the response.
-        :param function exempt_when: function/lambda used to decide if the rate
-         limit should skipped.
-        :param override_defaults:  whether the decorated limit overrides
-         the default limits. (default: True)
-        :param deduct_when: a function that receives the current
-         :class:`flask.Response`  object and returns True/False to decide if a
-         deduction should be done from the rate limit
-        """
-
-        return self.__limit_decorator(
-            limit_value,
-            key_func,
-            True,
-            scope,
-            error_message=error_message,
-            exempt_when=exempt_when,
-            override_defaults=override_defaults,
-            deduct_when=deduct_when,
-        )
-
-    def exempt(self, obj):
-        """
-        decorator to mark a view or all views in a blueprint as exempt from
-        rate limits.
-        """
-
-        if not isinstance(obj, Blueprint):
-            name = "%s.%s" % (obj.__module__, obj.__name__)
-
-            @wraps(obj)
-            def __inner(*a, **k):
-                return current_app.ensure_sync(obj)(*a, **k)
-
-            self._exempt_routes.add(name)
-
-            return __inner
-        else:
-            self._blueprint_exempt.add(obj.name)
-
-    def request_filter(self, fn: Callable) -> Callable:
-        """
-        decorator to mark a function as a filter to be executed
-        to check if the request is exempt from rate limiting.
-        """
-        self._request_filters.append(fn)
-
-        return fn
-
-    def __raise_global_limits_warning(self):
-        warnings.warn(
-            "global_limits was a badly named configuration since it is "
-            "actually a default limit and not a globally shared limit. Use "
-            "default_limits if you want to provide a default or use "
-            "application_limits if you intend to really have a global "
-            "shared limit",
-            UserWarning,
-        )
